import { isFunction, isUndefined } from './common-utils';

const RETRY_STATE_NOT_START = 0;
const RETRY_STATE_STARTED = 1;
const RETRY_STATE_STOPPED = 2;

/**
 * 给异步函数封装重试逻辑
 * @param {Object} options
 * @param {Object} options.retryFunction 需要封装的异步函数
 * @param {Object} options.settings
 * @param {Number} [options.settings.retries = 5] 重试次数
 * @param {Number} [options.settings.timeout = 1000] 重试间隔
 * @param {onErrorCallback} options.onError
 * @param {onRetryingCallback} [options.onRetrying]
 * @param {Object} options.context 上下文，可选
 * @return {Function} 封装后的函数
 * @example
 * const getUserMedia = promiseRetry({
 *   retryFunction: getUserMedia_,
 *   settings: { retries: 5, timeout: 2000 },
 *   onError: (error, retry, reject) => {
 *     if (error.name === 'NotReadableError') {
 *       retry();
 *     } else {
 *       reject(error);
 *     }
 *   },
 *   onRetrying: retryCount => {
 *     loggerManager.warn(`getUserMedia NotReadableError observed, retrying [${retryCount}/5]`);
 *   }
 * });
 */
function promiseRetry({ retryFunction, settings, onError, onRetrying, onRetryFailed, onRetrySuccess, context }) {
  return function(...args) {
    const retries = settings.retries || 5;
    let retryCount = 0;
    let timer = -1;
    let retryState = RETRY_STATE_NOT_START;
    const run = async (resolve, reject) => {
      const ctx = context || this;
      try {
        const result = await retryFunction.apply(ctx, args);
        if (retryCount !== 0) {
          if (isFunction(onRetrySuccess)) {
            onRetrySuccess.call(ctx, retryCount);
          }
        }
        // 执行成功，正常返回
        retryCount = 0;
        resolve(result);
      } catch (error) {
        // 用于停止重试
        const stopRetry = () => {
          clearTimeout(timer);
          retryCount = 0;
          retryState = RETRY_STATE_STOPPED;
          reject(error);
        };
        const retry = () => {
          if (retryState !== RETRY_STATE_STOPPED && retryCount < retries) {
            retryCount++;
            retryState = RETRY_STATE_STARTED;
            if (isFunction(onRetrying)) {
              onRetrying.call(ctx, retryCount, stopRetry);
            }
            timer = setTimeout(
              () => {
                timer = -1;
                run(resolve, reject);
              },
              isUndefined(settings.timeout) ? 1000 : settings.timeout
            );
          } else {
            stopRetry();
            if (isFunction(onRetryFailed)) {
              onRetryFailed.call(ctx, error);
            }
          }
        };
        if (isFunction(onError)) {
          onError.call(ctx, error, retry, reject, args, retryCount);
        } else {
          retry();
        }
      }
    };

    return new Promise(run);
  };
}

/**
 * @callback onErrorCallback
 * @param {Error} error 错误信息，可通过 error 来判断是否需要进行重试
 * @param {Function} retry 调用该回调执行重试
 * @param {Function} reject 当不需要重试的时候，reject(error)
 * @return undefined
 * @example
 * function onError(error, retry, reject, retryFuncArgs) {
 *    if (error.name === 'NotReadableError') {
 *      retry();
 *    } else {
 *      reject(error)
 *    }
 * }
 */

/**
 * @callback onRetryingCallback
 * @param {Number} retryCount 当前重试次数
 * @param {Function} stopRetry 可调用该回调停止重试
 * @return undefined
 */

export default promiseRetry;
